/*
 * NAME
 *      bc635.c - Cover functions for bc635pci driver devctl() commands.
 *
 * SYNOPSIS
 *
 *      #include "bc635.h"
 *
 *      int bc635_set_periodic_output(int fd, unsigned int period_usec, int one_pps_sync)
 *      int bc635_enable_int(int fd)
 *      int bc635_disable_int(int fd)
 *      int bc635_delay( int fd, unsigned int count)
 *      int bc635_set_mode(int fd, int mode)
 *      int bc635_set_time(int fd, int seconds)
 *      int bc635_get_time(int fd, double *seconds, int *lockstatus)
 *      int bc635_get_model(int fd, char *model, int len)
 *      int bc635_get_assembly(int fd, char *assembly, int len)
 *      int bc635_get_hwfab(int fd, char *hwfab, int len)
 *      int bc635_get_serial(int fd, char *serial, int len)
 *      int bc635_get_firmware_part(int fd, char *firmware_part, int len)
 *      int bc635_software_reset(int fd)
 *
 *      void bc635_lockstatus_str(int lockstatus, char *str)
 *      int bc635_open_s( void *exception( int en ))
 *      void bc635_timestamp_s( char *label )
 *      double bc635_get_time_s( void )
 *      void bc635_set_first_time_s( double time )
 *
 *
 * DESCRIPTION
 *
 *      These cover functions are used instead of raw devctl()
 *      calls to interact with the bc635pci time code processor card.
 *      For now, see comments in functions.
 *
 *      To use these, first call 'fd = open( "/dev/bc635pci", ...);', then use
 *      fd as first argument to these calls.
 *
 * AUTHOR
 *
 *      John H. Zouck
 *      The Johns Hopkins University
 *      Applied Physics Laboratory
 *      Johns Hopkins Rd.
 *      Laurel, MD 20723
 */

/*
 * DISCLAIMER:
 *
 *   THIS SOFTWARE IS SUPPLIED AS-IS BY THE JOHNS HOPKINS UNIVERSITY/APPLIED PHYSICS
 *   LABORATORY (JHU/APL).
 *
 *   WITH RESPECT TO THE INFORMATION, CODE AND/OR EXECUTABLES PROVIDED HEREIN,
 *   JHU/APL MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR IMPLIED,
 *   INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *   FITNESS FOR A PARTICULAR PURPOSE. IN ADDITION, IN NO EVENT SHALL JHU/APL
 *   BE HELD LIABLE FOR ANY DAMAGES WHATSOEVER INCLUDING DIRECT, INDIRECT, INCIDENTAL,
 *   CONSEQUENTIAL, LOSS OF BUSINESS PROFITS OR SPECIAL DAMAGES, EVEN IF
 *   JHU/APL HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/

#include <sys/types.h>
#include <sys/iofunc.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <devctl.h>
#include <errno.h>
#include <math.h>

#include "bc635dvr.h"
#include "bc635.h"




/*************  bc635_set_periodic_output() *************
 *
 * Set divisors for periodic output and interrupt (see page
 * 4 - 3 of bc635pci manual) at period of period_usec microseconds.
 *
 * Inputs
 *                      int fd - open file descriptor
 *    unsigned int period_usec - period in microseconds
 *            int one_pps_sync - Sync with 1 PPS epoch (BC635_SYNC_PPS),
 *                               or not (BC635_NO_SYNC_PPS)
 *
 * Returns:
 *
 *  BC635_FREQUENCY_RANGE_ERROR if exact divisors of period_usec
 *        cannot be found.
 *  BC635_DEVCTL_ERROR if devctl() is unsuccessful
 *  BC635_SUCCESS call is successful.
 *
 * Note: Syncing with 1 PPS will make periods that do not end on 1 second
 * epochs be cut short. For instance, if we set period of
 * 750,000 microseconds, then the second period each second will
 * be shortened, and the next will start on the 1 seconds epoch.
 *
 **********************************************************/

int
bc635_set_periodic_output(int fd, unsigned int period_usec, int one_pps_sync)
{
	uint16_t        divisors[2];
	unsigned int    n1;
	unsigned int    n2;
	int             info;
	unsigned int    srt;
	divisors_t      div;

	/*
	 * Need to get factors n1 and n2 such that n1 x n2 = period_usec
	 */

	/*
	 * To find factors for period_usec < 19,958,400, try numbers starting
	 * at 3 and going to 11
	 */

	if (period_usec > 19958400)
		return (BC635_FREQUENCY_RANGE_ERROR);

	/* Stop looking at sqrt( period_usec ) */

	srt = sqrt((double) period_usec);

	for (n1 = 3; n1 < srt; n1++) {
		n2 = period_usec / n1;
		if (n2 > 2 && n2 <= 65535 && n2 * n1 == period_usec)
			break;
	}

	if (n1 >= srt)
		return (BC635_FREQUENCY_RANGE_ERROR);

	div.divisors[1] = n2;	/* n2 */
	div.divisors[0] = n1;	/* n1 */

	div.one_pps_sync = one_pps_sync;

	/* Use devctl() to send buffer2 to device: */

	if (EOK != devctl(fd, BC635_SET_PERIODIC_OUTPUT, &div, sizeof(divisors_t), &info))
		return (BC635_DEVCTL_ERROR);

	return (BC635_SUCCESS);
}

/*************  bc635_enable_int() *************
 *
 * Enable interrupt
 *
 * Inputs
 *
 *  int fd - open file descriptor
 *
 * Returns:
 *
 *  BC635_DEVCTL_ERROR if devctl() is unsuccessful
 *  BC635_SUCCESS call is successful.
 *************************************************/

int
bc635_enable_int(int fd)
{
	int             info;
	if (EOK != devctl(fd, BC635_ENABLE_INT, 0, 0, &info))
		return (BC635_DEVCTL_ERROR);
	return (BC635_SUCCESS);
}


/*************  bc635_disable_int() *************
 *
 * Disable interrupt
 *
 * Inputs
 *
 *  int fd - open file descriptor
 *
 * Returns:
 *
 *  BC635_DEVCTL_ERROR if devctl() is unsuccessful
 *  BC635_SUCCESS call is successful.
 *************************************************/

int
bc635_disable_int(int fd)
{
	int             info;
	if (EOK != devctl(fd, BC635_DISABLE_INT, 0, 0, &info))
		return (BC635_DEVCTL_ERROR);
	return (BC635_SUCCESS);
}


/*************  bc635_delay() *************
 *
 * Delay a number of tics (periods)
 *
 * Inputs
 *
 *             int fd - open file descriptor
 * unsigned int count - number of tics to wait
 *
 * Returns:
 *
 *  BC635_DEVCTL_ERROR if devctl() is unsuccessful
 *  BC635_SUCCESS call is successful.
 *************************************************/

int
bc635_delay(int fd, unsigned int count)
{
	int             info;
	unsigned int    stat;

	stat = devctl(fd, BC635_DELAY, &count, sizeof( unsigned int), &info);
	if (EOK != stat)
		return (BC635_DEVCTL_ERROR);

	return (BC635_SUCCESS);
}


/*************  bc635_set_mode() *************
 *
 * Set bc635 major sync mode.
 *
 * Inputs
 *
 *  int fd - open file descriptor
 *  mode - one of DATUM_MODE_TIMECODE, DATUM_MODE_FREERUN,
 *         DATUM_MODE_EXT1PPS, DATUM_MODE_RTC, DATUM_MODE_GPS
 *
 * Returns:
 *
 *  BC635_DEVCTL_ERROR if devctl() is unsuccessful
 *  BC635_SUCCESS call is successful.
 *************************************************/

int
bc635_set_mode(int fd, int mode)
{
	int             info;
	if (EOK != devctl(fd, BC635_SET_MODE, &mode, sizeof(mode), &info))
		return (BC635_DEVCTL_ERROR);
	return (BC635_SUCCESS);
}



/*************  bc635_set_time() *************
 *
 * Set seconds on board clock for use
 * in modes other than
 * IRIG, where the time comes in on the signal. The
 * time only takes effect on the next 1 seconds epoch.
 *
 * Inputs
 *
 *       int fd - open file descriptor
 *  seconds - integral seconds.
 *
 * Returns:
 *
 *  BC635_DEVCTL_ERROR if devctl() is unsuccessful
 *  BC635_SUCCESS call is successful.
 *************************************************/


int
bc635_set_time(int fd, int seconds)
{
	int             info;
	if (EOK != devctl(fd, BC635_SET_TIME, &seconds, sizeof(seconds), &info))
		return (BC635_DEVCTL_ERROR);
	return (BC635_SUCCESS);
}

/*************  bc635_get_time() *************
 *
 * Get seconds from board clock
 *
 * Inputs
 *
 *           int fd - open file descriptor
 * Outputs
 *
 *  int *lockstatus - lock status return from card time field.
 *  double *seconds - integral seconds returned from card.
 *
 * Returns:
 *
 *  BC635_DEVCTL_ERROR if devctl() is unsuccessful
 *  BC635_SUCCESS call is successful.
 *************************************************/

int
bc635_get_time(int fd, double *seconds, int *lockstatus)
{
	time_struct_t   ts;
	int             info;

	if (EOK != devctl(fd, BC635_GET_TIME, &ts, sizeof(ts), &info))
		return (BC635_DEVCTL_ERROR);

	*seconds = ts.seconds;
	*lockstatus = ts.lockstatus;

	return (BC635_SUCCESS);
}


/************** bc635_get_model() **************
 *
 * Get board model number. Returns null-terminated string with
 * board model number.
 *
 * Inputs
 *
 *       int fd - open file descriptor
 *      int len - Maximum length of returned string.
 *
 * Outputs
 *
 *  char *model - printable model number returned.
 *
 * Returns:
 *
 *  BC635_DEVCTL_ERROR if devctl() is unsuccessful
 *  BC635_SUCCESS call is successful.
 *************************************************/

int
bc635_get_model(int fd, char *model, int len)
{
	int             info;

	if (EOK != devctl(fd, BC635_GET_MODEL, model, len, &info))
		return (BC635_DEVCTL_ERROR);
	return (BC635_SUCCESS);
}


/************** bc635_get_assembly() **************
 *
 * Get board assembly number. Returns null-terminated string with
 * board assembly number.
 *
 * Inputs
 *
 *       int fd - open file descriptor
 *      int len - Maximum length of returned string.
 *
 * Outputs
 *
 *  char *assembly - printable assembly number returned.
 *
 * Returns:
 *
 *  BC635_DEVCTL_ERROR if devctl() is unsuccessful
 *  BC635_SUCCESS call is successful.
 *************************************************/

int
bc635_get_assembly(int fd, char *assembly, int len)
{
	int             info;

	if (EOK != devctl(fd, BC635_GET_ASSEMBLY, assembly, len, &info))
		return (BC635_DEVCTL_ERROR);
	return (BC635_SUCCESS);
}


/************** bc635_get_hwfab() **************
 *
 * Get board hwfab number. Returns null-terminated string with
 * board hwfab number.
 *
 * Inputs
 *
 *       int fd - open file descriptor
 *      int len - Maximum length of returned string.
 *
 * Outputs
 *
 *  char *hwfab - printable hwfab number returned.
 *
 * Returns:
 *
 *  BC635_DEVCTL_ERROR if devctl() is unsuccessful
 *  BC635_SUCCESS call is successful.
 *************************************************/

int
bc635_get_hwfab(int fd, char *hwfab, int len)
{
	int             info;

	if (EOK != devctl(fd, BC635_GET_HWFAB, hwfab, len, &info))
		return (BC635_DEVCTL_ERROR);
	return (BC635_SUCCESS);
}




/************** bc635_get_serial() **************
 *
 * Get board serial number. Returns null-terminated string with
 * board serial number.
 *
 * Inputs
 *
 *       int fd - open file descriptor
 *      int len - Maximum length of returned string.
 *
 * Outputs
 *
 *  char *serial - printable serial number returned.
 *
 * Returns:
 *
 *  BC635_DEVCTL_ERROR if devctl() is unsuccessful
 *  BC635_SUCCESS call is successful.
 *************************************************/

int
bc635_get_serial(int fd, char *serial, int len)
{
	int             info;

	if (EOK != devctl(fd, BC635_GET_SERIAL, serial, len, &info))
		return (BC635_DEVCTL_ERROR);
	return (BC635_SUCCESS);
}


/************** bc635_get_firmware_part() **************
 *
 * Get firmware part number. Returns null-terminated string with
 * board firmware part.
 *
 * Inputs
 *
 *       int fd - open file descriptor
 *      int len - Maximum length of returned string.
 *
 * Outputs
 *
 *  char *firmware_part - printable board firmware partnumber returned.
 *
 * Returns:
 *
 *  BC635_DEVCTL_ERROR if devctl() is unsuccessful
 *  BC635_SUCCESS call is successful.
 *************************************************/

int
bc635_get_firmware_part(int fd, char *firmware_part, int len)
{
	int             info;

	if (EOK != devctl(fd, BC635_GET_FIRMWARE_PART, firmware_part, len, &info))
		return (BC635_DEVCTL_ERROR);
	return (BC635_SUCCESS);
}

/************** bc635_software_reset() **************
 *
 * Perform a software reset on board. Returns it to power on
 * configuration. It does not change any registers, however.
 *
 * Inputs
 *
 *       int fd - open file descriptor
 * Outputs
 *
 *  None
 *
 * Returns:
 *
 *  BC635_DEVCTL_ERROR if devctl() is unsuccessful
 *  BC635_SUCCESS call is successful.
 *************************************************/

int
bc635_software_reset(int fd)
{
	int             info;

	if (EOK != devctl(fd, BC635_SOFTWARE_RESET, 0, 0, &info))
		return (BC635_DEVCTL_ERROR);
	return (BC635_SUCCESS);
}



static char    *lockstatus_str[] = {
	"Locked to Selected Reference",
	"Not Locked to Selected Reference",
	"Time offset < X Microseconds",
	"Time offset >= X Microseconds",
	"Frequency offset < 5 x 10^-8 Hz",
"Frequency offset >= 5 x 10^-8 Hz"};

/*************  bc635_lockstatus_str() *************
 *
 * Convenience function. Returns string with description of lockstatus
 *
 * Inputs
 *
 *       int lockstatus - lockstatus return from bc635_gettime().
 * Outputs
 *
 *  char *str - Printable interpretation of lockstatus.
 *
 * Returns:
 *
 *  None.
 *************************************************/

void
bc635_lockstatus_str(int lockstatus, char *str)
{
	sprintf(str, "%s, %s, %s.",
		lockstatus_str[lockstatus & 1],
		lockstatus_str[((lockstatus >> 1) & 1) + 2],
		lockstatus_str[((lockstatus >> 2) & 1) + 4]);
}
