/*
 * NAME
 *     bc635pci.c - Bancomm bc635pci resource manager for QRTP
 *
 * DESCRIPTION
 *
 *     Searches PCI bus for bc635pci boards and initializes
 *     first board found. Creates /dev/bc635pci device for
 *     application programs to open and access board.
 *
 *     Initializes board and handles devctl() calls to peform
 *     functions for application programs, including:
 *
 *     - Read time in seconds double precision
 *     - Read part number, serial number, software revision and date
 *
 *     - Set major time in integer seconds
 *     - Set board time sync mode: external time code, free run, external 1 PPS,
 *       real-time clock.
 *     - Set board to output periodic signal and interrupt periodically
 *
 *     - Wait for periodic interrupt, which allows application threads
 *       to be woken at precise periods
 *
 *     Implements client (application) calls:
 *
 *        open(), close(), devctl()
 *
 * USE
 *
 *     To start resource manager and register name bc635pci in /dev:
 *
 *        # bc635pci&
 *
 *     devctl() calls:
 *     ---------------
 *
 *     Note: applications would normally use the library of convenient
 *     cover functions to perform these devctl() calls.
 *
 *     TBD when baseline complete.
 *
 * SEE ALSO
 *
 *     bc635.c devctl() cover functions.
 *
 * NOTES/TO-DO
 *
 *     This resource manager is incomplete. It implements a core of functions to
 *     allow the card to be used for basic timing applications.
 *
 *     Framework from http://qdn.qnx.com/support/docs/neutrino_qrp/prog/resmgr.html
 *
 * AUTHOR
 *
 * John H. Zouck
 * The Johns Hopkins University
 * Applied Physics Laboratory
 * Johns Hopkins Rd.
 * Laurel, MD 20723
 *
 */

/*
 * DISCLAIMER:
 *
 *   THIS SOFTWARE IS SUPPLIED AS-IS BY THE JOHNS HOPKINS UNIVERSITY/APPLIED PHYSICS
 *   LABORATORY (JHU/APL).
 *
 *   WITH RESPECT TO THE INFORMATION, CODE AND/OR EXECUTABLES PROVIDED HEREIN,
 *   JHU/APL MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR IMPLIED,
 *   INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *   FITNESS FOR A PARTICULAR PURPOSE. IN ADDITION, IN NO EVENT SHALL JHU/APL
 *   BE HELD LIABLE FOR ANY DAMAGES WHATSOEVER INCLUDING DIRECT, INDIRECT, INCIDENTAL,
 *   CONSEQUENTIAL, LOSS OF BUSINESS PROFITS OR SPECIAL DAMAGES, EVEN IF
 *   JHU/APL HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/

/********************** Includes **********************/

#include <hw/pci.h>
#include <sys/mman.h>
#include <sys/neutrino.h>
#include <sys/siginfo.h>
#include <errno.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/iofunc.h>
#include <sys/dispatch.h>
#include <string.h>
#include <devctl.h>
#include <pthread.h>

#include "bc635dvr.h"
#include "bc635.h"
#include "utils.h"

/********************** Defines **********************/

/* Vendor ID for Datum Bancomm bc635pci card */

#define BC635PCI_VENDORID ( 0x12e2 )

/* Device ID  for Datum Bancomm bc635pci card */

#define BC635PCI_DEVICEID ( 0x4013 )

/* Maximum number of waiting threads or processes */

#define NWAITMAX 100


#define DATUM_DPRAM_TOP         0x800

/* PCI data */

#define DATUM_PCI_VENDOR_ID     0x12E2
#define DATUM_PCI_DEVICE_ID     0x4013
#define DATUM_DPRAM_TOP         0x800

/* Device register control field bit definitions */

#define DATUM_CONTROL_BIT_LOCKEN        0x01
#define DATUM_CONTROL_BIT_EVSOURCE      0x02
#define DATUM_CONTROL_BIT_EVSENSE       0x04
#define DATUM_CONTROL_BIT_EVENTEN       0x08
#define DATUM_CONTROL_BIT_STREN         0x10
#define DATUM_CONTROL_BIT_STRMODE       0x20
#define DATUM_CONTROL_BIT_FREQSEL0      0x40
#define DATUM_CONTROL_BIT_FREQSEL1      0x80

/* Device register interrupt mask and status field bit definitions */

#define DATUM_INT_BIT_EVENT             0x01
#define DATUM_INT_BIT_PERIOD    0x02
#define DATUM_INT_BIT_STROBE    0x04
#define DATUM_INT_BIT_1PPS              0x08
#define DATUM_INT_BIT_GPS               0x10
#define DATUM_INT_BIT_ALL               ( DATUM_INT_BIT_EVENT |\
                                          DATUM_INT_BIT_PERIOD|\
                                          DATUM_INT_BIT_STROBE|\
                                          DATUM_INT_BIT_1PPS  |\
                                          DATUM_INT_BIT_GPS )

/* Time data status bit field definitions */

#define DATUM_STATUS_BIT_FLY    0x01
#define DATUM_STATUS_BIT_TOFS   0x02
#define DATUM_STATUS_BIT_FOFS   0x04

/* Device register acknowledge field bit definitions */

#define DATUM_ACK_BIT_REC               0x01
#define DATUM_ACK_BIT_GPS               0x02
#define DATUM_ACK_BIT_CMD               0x80

/* Dual port ram commands */

#define DATUM_SET_TIMING_MODE_CMD       0x10
#define DATUM_MODE_TIMECODE             0x00
#define DATUM_MODE_FREERUN              0x01
#define DATUM_MODE_EXT1PPS              0x02
#define DATUM_MODE_RTC                  0x03
#define DATUM_MODE_GPS                  0x06

#define DATUM_SET_TIME_FORMAT_CMD       0x11
#define DATUM_TIME_FMT_BCD              0x00
#define DATUM_TIME_FMT_UNIX             0x01

#define DATUM_SET_TIME_CMD                      0x12

#define DATUM_SET_YEAR_CMD                      0x13

#define DATUM_SET_PERIODIC_CMD          0x14
#define DATUM_NO_SYNC_PPS               0x00
#define DATUM_SYNC_PPS                  0x01

#define DATUM_SET_TC_FORMAT_CMD         0x15
#define DATUM_TIME_CODE_IRIGA   0x41
#define DATUM_TIME_CODE_IRIGB   0x42
#define DATUM_TIME_CODE_1344            0x49

#define DATUM_SET_TC_MOD_CMD            0x16
#define DATUM_MODE_AMP                  0x4D
#define DATUM_MODE_PULSE                        0x44

#define DATUM_SET_TIMING_OFFSET_CMD     0x17

#define DATUM_REQ_UTC_TM_DATA_CMD       0x18
#define DATUM_GPS_TIME_UTC              0x00
#define DATUM_GPS_TIME_GPS              0x01
#define DATUM_LEAP_DELETE               0xFF
#define DATUM_LEAP_NONE                 0x00
#define DATUM_LEAP_ADD                  0x01

#define DATUM_GET_DATA_CMD                      0x19
#define DATUM_GET_DATA_MODE             0x10
#define DATUM_GET_DATA_TFORMAT  0x11
#define DATUM_GET_DATA_YEAR             0x13
#define DATUM_GET_DATA_PER              0x14
#define DATUM_GET_DATA_TCFMT            0x15
#define DATUM_GET_DATA_TCMOD            0x16
#define DATUM_GET_DATA_TOFF             0x17
#define DATUM_GET_DATA_UTCINFO  0x18

#define DATUM_GET_DATA_TCOUTFMT 0x1B
#define DATUM_GET_DATA_TCGENOFF 0x1C
#define DATUM_GET_DATA_LOCTMOFF 0x1D
#define DATUM_GET_DATA_LEAPSEC  0x1E

#define DATUM_GET_DATA_FWVER            0x1F
#define DATUM_GET_DATA_CLKSRC   0x20
#define DATUM_GET_DATA_JAMSC            0x21
#define DATUM_GET_DATA_OSCCTL   0x23
#define DATUM_GET_DATA_DAVAL            0x24

#define DATUM_GET_DATA_BATTSTAT 0x26
#define DATUM_GET_DATA_CLKVAL   0x29
#define DATUM_GET_DATA_DTFW             0x4F
#define DATUM_GET_DATA_ASMB             0xF4
#define DATUM_GET_DATA_HWFAB            0xF5
#define DATUM_GET_DATA_TFPMODEL 0xF6
#define DATUM_GET_DATA_SERIAL   0xFE

#define DATUM_SOFT_RESET_CMD            0x1A

#define DATUM_SET_TC_OUT_FMT_CMD        0x1B

#define DATUM_SET_GEN_TM_OFF_CMD        0x1C
#define DATUM_HALF_NO                   0x00
#define DATUM_HALF_YES                  0x01

#define DATUM_SET_LOC_TM_OFF_CMD        0x1D

#define DATUM_SET_LEAP_SEC_CMD          0x1E

#define DATUM_SEL_CLOCK_CMD                     0x20
#define DATUM_SEL_CLOCK_INT             0x49
#define DATUM_SEL_CLOCK_EXT             0x45

#define DATUM_JAM_SYNC_ENABLE_CMD       0x21
#define DATUM_JAMSYNC_DIS               0x00
#define DATUM_JAMSYNC_EN                        0x01

#define DATUM_FORCEJAMSYNC_CMD          0x22

#define DATUM_DISC_ENABLE_CMD           0x23
#define DATUM_DISC_DIS                  0x00
#define DATUM_DISC_EN                   0x01

#define DATUM_SET_DAC_CMD                       0x24

#define DATUM_SET_DISC_GAIN_CMD         0x25

#define DATUM_REQ_BATT_STAT_CMD         0x26
#define DATUM_BATT_FAILED               0x00
#define DATUM_BATT_OK                   0x10

#define DATUM_SYNC_RTC_CMD                      0x27

#define DATUM_DISC_BATT_RTC_CMD         0x28

#define DATUM_SET_CLK_VAL_CMD           0x29

#define DATUM_SEND_GPS_PKT_CMD          0x30

#define DATUM_REQ_GPS_PKT_CMD           0x31

#define DATUM_MAN_REQ_GPS_PKT_CMD       0x32

#define DATUM_GPS_TIME_FORMAT_CMD       0x33

#define DATUM_SET_GPS_MD_FLG_CMD        0x34
#define DATUM_DISABLE                   0x00
#define DATUM_ENABLE                            0x01

#define DATUM_SET_LOC_TM_FLG_CMD        0x40

#define DATUM_REQ_DAYLT_FLG_CMD         0x41
#define DATUM_LOCAL_ENABLE              0x08

#define DATUM_SET_YR_INC_FLG_CMD        0x42

#define DATUM_REQ_DTFW_NUM_CMD          0x4F
#define DATUM_ID_DTFW                   "DT6000"

#define DATUM_REQ_ASSEMB_CMD            0xF4
#define DATUM_OLD_ASM_PART              12043
#define DATUM_NEW_ASM_PART              12083

#define DATUM_REQ_HW_FAB_CMD            0xF5
#define DATUM_OLD_FAB_PART              12042
#define DATUM_NEW_FAB_PART              12083

#define DATUM_REQ_TFP_MODEL_CMD         0xF6
#define DATUM_ID_IRIG                   "BC635PCI"
#define DATUM_ID_GPS                            "BC637PCI"

#define DATUM_REQ_SER_NUM_CMD           0xFE

/* Max length of messages to bc635 in dpram */

#define DATUM_MAX_CMD_LEN		100

/* Max length of messages from bc635 in dpram */

#define DATUM_MAX_OUT_MSG_LEN		100

/* Structure to overlay bc635pci card registers */

typedef struct {
	unsigned long   timereq;
	unsigned long   eventreq;
	unsigned long   unlock;
	unsigned long   r1;
	unsigned long   control;
	unsigned long   ack;
	unsigned long   mask;
	unsigned long   intstat;
	unsigned long   minstrb;
	unsigned long   majstrb;
	unsigned long   r2;
	unsigned long   r3;
	unsigned long   time0;
	unsigned long   time1;
	unsigned long   event0;
	unsigned long   event1;
}               registers_t;

/*
 * Define structure to hold almost all static global data
 * We may need this if we support multiple boards at some time.
 */

/* Structure to hold rcvids waiting for reply */

typedef struct {
	int id;
	unsigned int count_remaining;
} bc635_rcvid_t;

typedef struct {
	pthread_mutex_t mut;
	bc635_rcvid_t rcvids[NWAITMAX];
}               bc635_wait_t;

typedef struct {
	uint8_t        *dpram_base;
	volatile registers_t *register_base;
	struct pci_dev_info info;
	int             phdl;
	void           *hdl;
	char           *dpram_in_area;
	char           *dpram_out_area;
	char           *dpram_year_area;
	bc635_wait_t    bc635_waiters;
	uint32_t        intno;
	/* Interrupt line number */
	uint16_t divisors[ 2 ];
}               pstat_t;

/*** Global static structure with per-board information ***/

static pstat_t  glbl;

/****************** Local function Prototypes ******************/

static int      io_devctl(resmgr_context_t * ctp, io_devctl_t * msg, RESMGR_OCB_T * ocb);
static int      io_close(resmgr_context_t * ctp, void *reserved, RESMGR_OCB_T * ocb);
static int      scan_bc635_cards(pstat_t * glbl);
static void     read_time(time_struct_t * ts);
static int      bcmd_set_periodic_output(divisors_t * div);
static void     bcmd_set_major_time(uint32_t seconds);
static int      bcmd_set_timing_mode(uint8_t mode);
static int      bcmd_get_model(char *model);
static int      bcmd_get_hwfab(char *hwfab);
static int      bcmd_get_firmware_part(char *firmware_part);
static int      bcmd_get_assembly(char *assembly);
static int      bcmd_get_serial(char *serial);
static int      bcmd_software_reset(void);
static void     write_dpram_data(char *data, int len);
static void     read_dpram_data(char *data, int offset, int len);
static int      isr_settup(void);
static void     bc635_waiters_reply(int reply);
static int      bc635_waiters_add(int rcvid, unsigned int count);
static void     bc635_waiters_init(void);

/****************** Local Static Data Declarations ******************/

static resmgr_connect_funcs_t connect_funcs;
static resmgr_io_funcs_t io_funcs;
static iofunc_attr_t attr;



/****************** main() ******************
 *
 * This resource manager is, like all QRTP resource managers, simply
 * a program run as a process, so it has a main() function.
 *
 */

main(int argc, char **argv)
{
	/* declare variables we 'll be using */

	dispatch_t     *dpp;
	resmgr_attr_t   resmgr_attr;
	resmgr_context_t *ctp;
	int             id;
	char            model[15];
	char            firmware_part[15];
	char            serial[15];

	printf("bc635pci time code processor resource manager.\n");

	/* TBD:	Set up bc635 board PCI stuff. Initialize board. */

	if (EXIT_FAILURE == scan_bc635_cards(&glbl)) {
		printf("bc635pci resource manager found no valid bc635pci cards.\n");
		exit(1);
	}
	/* do software reset of card. */

	printf("Doing software reset of bc635pci card.\n");

	bcmd_software_reset();

	/* Delay while card initializes */

	sleep(3);

	/* Get some stuff to print out on startup. */

	bcmd_get_model(model);
	bcmd_get_firmware_part(firmware_part);
	bcmd_get_serial(serial);

	printf("Found bc635 card.\n");
	printf("Model: %s Serial Number: %s Firmware Version: %s\n",
	       model, serial, firmware_part);

	/* Start up interrupt system */

	isr_settup();

	/* initialize dispatch interface */

	if ((dpp = dispatch_create()) == NULL) {
		fprintf(stderr, "%s: Unable to allocate dispatch handle.\n",
			argv[0]);
		return EXIT_FAILURE;
	}
	/* initialize resource manager attributes */

	memset(&resmgr_attr, 0, sizeof resmgr_attr);
	resmgr_attr.nparts_max = 1;
	resmgr_attr.msg_max_size = 2048;

	/* initialize functions for handling messages */

	iofunc_func_init(_RESMGR_CONNECT_NFUNCS, &connect_funcs, _RESMGR_IO_NFUNCS,
			 &io_funcs);

	/* All functions are accessed through devctl() */

	io_funcs.devctl = io_devctl;
	io_funcs.close_ocb = io_close;

	/* initialize attribute structure */

	iofunc_attr_init(&attr, S_IFNAM | 0666, 0, 0);

	/*
	 * Set attr.nbytes to 0 because this device has no natural length,
	 * like a file might.
	 */

	attr.nbytes = 0;

	/* attach our device name */

	if ((id = resmgr_attach(dpp, &resmgr_attr, "/dev/bc635pci", _FTYPE_ANY, 0,
				&connect_funcs, &io_funcs, &attr)) == -1) {
		fprintf(stderr, "%s: Unable to attach name.\n", argv[0]);
		return EXIT_FAILURE;
	}
	/* allocate a context structure */

	ctp = resmgr_context_alloc(dpp);

	/* start the resource manager message loop */

	while (1) {
		if ((ctp = resmgr_block(ctp)) == NULL) {
			fprintf(stderr, "block error\n");
			return EXIT_FAILURE;
		}
		resmgr_handler(ctp);
	}
}


/********************** io_devctl() **********************
 *
 * This function handles application calls to devctl(). Library bc635.c
 * provides friendlier cover functions for these devctl() calls.
 */


static int
io_devctl(resmgr_context_t * ctp, io_devctl_t * msg, RESMGR_OCB_T * ocb)
{
	int             cmd;
	int             dir;
	int             status;
	int             mode;
	int             seconds;

	status = EOK;

	cmd = get_device_command(msg->i.dcmd);
	dir = get_device_direction(msg->i.dcmd);

	switch (cmd) {

	case get_device_command(BC635_SET_PERIODIC_OUTPUT):

		/*
		 * This function sets divisors and starts periodic output,
		 * which can also cause interrupts at the
		 * same rate as the output, if they are enabled.
		 *
		 */

		/*
		 * printf("bc635pci command %s received.\n",
		 * "BC635_SET_PERIODIC_OUTPUT");
		 */
		/* Get divisors */
		if (dir == DEVDIR_TO) {
			divisors_t      div;
			resmgr_msgread(ctp, &div, msg->i.nbytes, sizeof(msg->i));

			/* Setup bc635 periodic output. */

			if (BC635_SUCCESS != bcmd_set_periodic_output(&div))
				status = ENOTTY;
		} else {
			printf("Wrong data direction on devctl()\n");
			status = ENOTTY;
		}
		break;


	case get_device_command(BC635_ENABLE_INT):

		/*
		 * Enable periodic interrupts on board by setting interrupt mask
		 * register.
		 */

		/*
		 * printf("bc635pci command %s received.\n",
		 * "BC635_ENABLE_INT");
		 */

		/* Enable periodic interrupt */

		glbl.register_base->mask = DATUM_INT_BIT_PERIOD;

		break;

	case get_device_command(BC635_DISABLE_INT):

		/*
		 * Disable interrupts on board by setting interrupt mask
		 * register to 0.
		 */

		/*
		 * printf("bc635pci command %s received.\n",
		 * "BC635_DISABLE_INT");
		 */

		/* Disable periodic interrupt */

		glbl.register_base->mask = 0x0;

		/* Release any waiters */

		bc635_waiters_reply(ENOTTY);

		break;

	case get_device_command(BC635_DELAY):

		/*
		 *  Delay for count periodic 'tics'
		 */

		/* printf("bc635pci command %s received.\n", "BC635_DELAY"); */

		/* Do not allow to delay if interrupts not enabled. */

		if ((glbl.register_base->mask & 0x2) == 0) {
			status = ENOTTY;
		} else if (dir == DEVDIR_TO) {
			unsigned int count;

			/* Get count to delay */

			resmgr_msgread(ctp, &count, msg->i.nbytes, sizeof(msg->i));

			/* Save reply rcvid for interrupt program to sendto. */

			if ( EXIT_FAILURE == bc635_waiters_add(ctp->rcvid, count) ) {
				status = ENOTTY;
			} else {

				/* Do not reply yet. */

				status = _RESMGR_NOREPLY;
			}
		} else {
			status = ENOTTY;
			printf("Wrong data direction on devctl()\n");
		}
		break;

	case get_device_command(BC635_SOFTWARE_RESET):

		/*
		 * Perform software reset on board. Don't know exactly
		 * what this does at this time. It does not set any
		 * registers.
		 */

		/*
		 * printf("bc635pci command %s received.\n",
		 * "BC635_SOFTWARE_RESET");
		 */
		bcmd_software_reset();
		break;

	case get_device_command(BC635_SET_MODE):

		/*
		 * Set synchronization mode of board.
		 */

		/* printf("bc635pci command %s received.\n", "BC635SETMODE"); */
		if (dir == DEVDIR_TO) {
			resmgr_msgread(ctp, &mode, msg->i.nbytes, sizeof(msg->i));
			if (bcmd_set_timing_mode((uint8_t) mode))
				status = ENOTTY;
		} else {
			status = ENOTTY;
			printf("Wrong data direction on devctl()\n");
		}
		break;

	case get_device_command(BC635_SET_TIME):

		/*
		 * Set major time (seconds) on board.
		 */

		/* printf("bc635pci command %s received.\n", "BC635SETTIME"); */
		if (dir == DEVDIR_TO) {
			resmgr_msgread(ctp, &seconds, msg->i.nbytes, sizeof(msg->i));
			/* Place code to set bc635 card time here. */
			bcmd_set_major_time(seconds);
		} else {
			status = ENOTTY;
			printf("Wrong data direction on devctl()\n");
		}
		break;

	case get_device_command(BC635_GET_TIME):

		/*
		 *  Return time and lock status
		 */

		if (dir == DEVDIR_FROM) {
			time_struct_t   ts;

			read_time(&ts);

			resmgr_msgwrite(ctp, &ts, msg->i.nbytes, sizeof(msg->i));
		} else {
			status = ENOTTY;
			printf("Wrong data direction on devctl()\n");
		}
		break;


	case get_device_command(BC635_GET_MODEL):

		/*
		 *  Get model number as string
		 */

		/*
		 * printf("bc635pci command %s received.\n",
		 * "BC635_GET_MODEL");
		 */
		if (dir == DEVDIR_FROM) {
			char            model[DATUM_MAX_OUT_MSG_LEN] = "NAS";
			int             len;

			bcmd_get_model(model);
			len = min(msg->i.nbytes, 1 + strlen(model));
			resmgr_msgwrite(ctp, model, len, sizeof(msg->i));
		} else {
			status = ENOTTY;
			printf("Wrong data direction on devctl()\n");
		}
		break;

	case get_device_command(BC635_GET_ASSEMBLY):

		/*
		 *  Get assembly number as string
		 */

		if (dir == DEVDIR_FROM) {
			char            assembly[DATUM_MAX_OUT_MSG_LEN] = "";
			int             len;

			bcmd_get_assembly( assembly );
			len = min(msg->i.nbytes, 1 + strlen(assembly));
			resmgr_msgwrite(ctp, assembly, len, sizeof(msg->i));
		} else {
			status = ENOTTY;
			printf("Wrong data direction on devctl()\n");
		}
		break;

	case get_device_command(BC635_GET_HWFAB):

		/*
		 *  Get hardware fab number as string
		 */

		if (dir == DEVDIR_FROM) {
			char            hwfab[DATUM_MAX_OUT_MSG_LEN] = "";
			int             len;

			bcmd_get_hwfab( hwfab );
			len = min(msg->i.nbytes, 1 + strlen(hwfab));
			resmgr_msgwrite(ctp, hwfab, len, sizeof(msg->i));
		} else {
			status = ENOTTY;
			printf("Wrong data direction on devctl()\n");
		}
		break;

	case get_device_command(BC635_GET_SERIAL):

		/*
		 *  Get serial number as string
		 */

		/*
		 * printf("bc635pci command %s received.\n",
		 * "BC635_GET_SERIAL");
		 */
		if (dir == DEVDIR_FROM) {
			char            serial[DATUM_MAX_OUT_MSG_LEN] = "";
			int             len;

			bcmd_get_serial(serial);
			len = min(msg->i.nbytes, 1 + strlen(serial));
			resmgr_msgwrite(ctp, serial, len, sizeof(msg->i));
		} else {
			status = ENOTTY;
			printf("Wrong data direction on devctl()\n");
		}
		break;

	case get_device_command(BC635_GET_FIRMWARE_PART):

		/*
		 *  Get firmware number as string
		 */

		/*
		 * printf("bc635pci command %s received.\n",
		 * "BC635_GET_FIRMWARE_PART");
		 */
		if (dir == DEVDIR_FROM) {
			char            firmware_part[DATUM_MAX_OUT_MSG_LEN] = "";
			int             len;

			bcmd_get_firmware_part(firmware_part);
			len = min(msg->i.nbytes, 1 + strlen(firmware_part));
			resmgr_msgwrite(ctp, firmware_part, len, sizeof(msg->i));
		} else {
			status = ENOTTY;
			printf("Wrong data direction on devctl()\n");
		}
		break;

	default:
		status = ENOTTY;
		printf("Unknown command received: 0x%08x\n", cmd);
	}

	return (status);
}

/******************** io_close() *****************
 *
 * If last close(), turn off interrupts so board does not
 * load processor unnecessarily. (This function is only called
 * when last close() has been run.)
 *
 */

static int
io_close(resmgr_context_t * ctp, void *reserved, RESMGR_OCB_T * ocb)
{

/*
	printf("io_close() called. count = %d\n",
	       ocb->attr->count);
*/

	/*
	 * Disable interrupts if count == 1. We test for 1 since we have not
	 * yet called iofunc_close_ocb_default(), which decrements the open count.
	 */

	if (ocb->attr->count == 1)
		glbl.register_base->mask = 0;

	bc635_waiters_reply(ENOTTY);

	return (iofunc_close_ocb_default(ctp, reserved, ocb));
}

/******************** scan_bc635_cards() ******************
 *
 * Look for 1st bc635pci PCI card in system and map it to process address
 * space so we can manupluate the registers and dual-port ram (dpram).
 *
 */

static int
scan_bc635_cards(pstat_t * glbl)
{

	uint16_t        pidx;

	/* The obligatory pci_attach() to talk to the PCI BIOS */

	glbl->phdl = pci_attach(0);
	if (glbl->phdl == -1) {
		perror("pci_attach");
		return (EXIT_FAILURE);
	}
	/* Clear out info */

	memset(&glbl->info, 0, sizeof(glbl->info));

	/* Device index: which PCI card with this vendor and device id */

	pidx = 0;

	/*
	 * Look up Bancomm bc635pci by vendor and device
	 */

	glbl->info.VendorId = BC635PCI_VENDORID;
	glbl->info.DeviceId = BC635PCI_DEVICEID;

	/*
	 * Do pci_attach_device() to find device, and create handle
	 */

	glbl->hdl = pci_attach_device(NULL, PCI_SHARE | PCI_SEARCH_VENDEV | PCI_INIT_ALL,
				      pidx, &glbl->info);

	if (glbl->hdl == NULL) {
		fprintf(stderr, "pci_attach_device() failed. No bc635pci boards found.\n");
		return (EXIT_FAILURE);
	}

	printf("First bc635pci card found. Details:\n");

	pciinfoshowhex( &glbl->info );

	/*
	 * Map memory segment 0 to process: register space
	 * (Documents are wrong: registers are at base 0, not
	 * dpram)
	 */

	glbl->register_base = mmap_device_memory(NULL, glbl->info.BaseAddressSize[0],
			    PROT_WRITE | PROT_READ | PROT_NOCACHE, MAP_TYPE,
					      glbl->info.CpuBaseAddress[0]);

	/*
	 * Map memory segment 1 to process: dpram space
	 */

	glbl->dpram_base = mmap_device_memory(NULL, glbl->info.BaseAddressSize[1],
			    PROT_WRITE | PROT_READ | PROT_NOCACHE, MAP_TYPE,
					      glbl->info.CpuBaseAddress[1]);

	if (glbl->dpram_base == MAP_FAILED || glbl->register_base == MAP_FAILED) {
		printf("mmap_device_memory() failed for base 0 or 1.\n");
		pci_detach_device(glbl->hdl);
		pci_detach(glbl->phdl);
		return (EXIT_FAILURE);
	}
	/* Set in, out area addresses. */

	glbl->dpram_in_area = glbl->dpram_base + *(glbl->dpram_base +
						   DATUM_DPRAM_TOP - 1) + *(glbl->dpram_base + DATUM_DPRAM_TOP - 2) * 256;

	glbl->dpram_out_area = glbl->dpram_base + *(glbl->dpram_base +
						    DATUM_DPRAM_TOP - 3) + *(glbl->dpram_base + DATUM_DPRAM_TOP - 4) * 256;

	/* Initialize registers */

	glbl->register_base->mask = 0;
	glbl->register_base->control = 0;
	glbl->register_base->intstat = 0;

	/* Save interrupt number for InterruptAttach() */

	glbl->intno = glbl->info.Irq;

	printf("bc635pci IRQ = %d\n", glbl->intno);

	return (EXIT_SUCCESS);
}

/********************** read_time() **********************
 *
 * Get time from card registers and convert to double type.
 * Latch first by accessing glbl.register_base->timereq.
 * Return time and lock status.
 *
 */

static void
read_time(time_struct_t * ts)
{
	unsigned int    seconds;
	unsigned int    useconds;
	unsigned int    strobe;

	/* Must latch time */

	strobe = glbl.register_base->timereq;

	seconds = glbl.register_base->time1;
	useconds = glbl.register_base->time0;
	ts->lockstatus = (useconds >> 24) & 0xf;
	useconds = useconds & 0xfffff;

	/* Convert to double */

	ts->seconds = (double) seconds + ((double) useconds) / 1000000.0;
}

/************* Functions that Issue bc635pci Commands via dpram **********/

/********** bcmd_set_major_time( uint32_t seconds ) **********
*
* Write set major time message command to set seconds. Will
* take effect next 1 pps epoch. User should issue this at
* least 200 ms before next 1 sec epoch.
*
*/

static void
bcmd_set_major_time(uint32_t seconds)
{
	uint8_t         buffer[DATUM_MAX_CMD_LEN];
	uint32_t        sec;

	/* Set command byte */

	buffer[0] = DATUM_SET_TIME_CMD;

	/*
	 * Byte ordering bc635 wants in network byte ordering, so convert
	 * from host byte ordering.
	 */

	sec = htonl(seconds);

	/* Copy seconds to buffer */

	memcpy(&buffer[1], (char *) &sec, sizeof(sec));

	write_dpram_data(buffer, 5);
}

/***************** bcmd_set_periodic_output() ****************
*
* Set divisors and enable periodic output.
*
*/

static int
bcmd_set_periodic_output(divisors_t * div)
{
	uint8_t         buffer[DATUM_MAX_CMD_LEN];
	uint16_t        n1;
	uint16_t        n2;

	/* To work right n1 and n2 must both be > 2 */

	if (div->divisors[0] <= 2 || div->divisors[1] <= 2)
		return (BC635_FREQUENCY_RANGE_ERROR);

	/* Save these for reference and return later. */

	glbl.divisors[ 0 ] = div->divisors[0];
	glbl.divisors[ 1 ] = div->divisors[1];

	buffer[0] = DATUM_SET_PERIODIC_CMD;

	/* Put in right byte order */

	n1 = htons(div->divisors[0]);
	n2 = htons(div->divisors[1]);

	/*
	 * Set sync with 1 PPS field.
	 *
	 * Note: If set to 1, this will make periods that do not end on 1 second
	 * epochs be cut short. For instance, if we set period of
	 * 750,000 microseconds, then the second period each second will
	 * be shortened, and th enext will start on the 1 seconds epoch.
	 */

	buffer[1] = div->one_pps_sync;

	memcpy(&buffer[2], (char *) &n1, 2);
	memcpy(&buffer[4], (char *) &n2, 2);

	write_dpram_data(buffer, 6);

	return (BC635_SUCCESS);
}


/**************** bcmd_set_timing_mode() ***************
*
* Issue command to set timing mode.
*/

static int
bcmd_set_timing_mode(uint8_t mode)
{
	uint8_t         buffer[DATUM_MAX_CMD_LEN];

	/*
	 * Check mode.
	 */

	switch (mode) {
	case DATUM_MODE_TIMECODE:
	case DATUM_MODE_FREERUN:
	case DATUM_MODE_EXT1PPS:
	case DATUM_MODE_RTC:
	case DATUM_MODE_GPS:

		buffer[0] = DATUM_SET_TIMING_MODE_CMD;
		buffer[1] = mode;
		write_dpram_data(buffer, 2);
		break;

	default:

		return (EXIT_FAILURE);
	}
	return (EXIT_SUCCESS);
}

/******************* bcmd_get_model() ******************
*
* Issue command to get board model number.
*
*/

static int
bcmd_get_model(char *model)
{
	uint8_t         buffer[DATUM_MAX_CMD_LEN];

	buffer[0] = DATUM_GET_DATA_CMD;
	buffer[1] = DATUM_REQ_TFP_MODEL_CMD;
	write_dpram_data(buffer, 2);
	read_dpram_data(buffer, 0, 1);
	if (buffer[0] != DATUM_REQ_TFP_MODEL_CMD)
		return (EXIT_FAILURE);

	/* Read rest of data */

	read_dpram_data(model, 1, 8);
	model[8] = 0;

	return (EXIT_SUCCESS);
}


/******************* bcmd_get_assembly() ******************
*
* Issue command to get board assembly number.
*
*/

static int
bcmd_get_assembly(char *assembly)
{
	uint8_t         buffer[DATUM_MAX_CMD_LEN];
	uint16_t        assembly_number;


	buffer[0] = DATUM_GET_DATA_CMD;
	buffer[1] = DATUM_REQ_ASSEMB_CMD;

	write_dpram_data(buffer, 2);

	read_dpram_data(buffer, 0, 1);

	if (buffer[0] != DATUM_REQ_ASSEMB_CMD)
		return (EXIT_FAILURE);

	read_dpram_data((char *) &assembly_number, 1, 2);

	/* Put in host order */

	assembly_number = ntohs(assembly_number);

	sprintf(assembly, "%d", assembly_number);

	return (EXIT_SUCCESS);
}


/******************* bcmd_get_hwfab() ******************
*
* Issue command to get board hwfab number.
*
*/

static int
bcmd_get_hwfab(char *hwfab)
{
	uint8_t         buffer[DATUM_MAX_CMD_LEN];
	uint16_t        hwfab_number;


	buffer[0] = DATUM_GET_DATA_CMD;
	buffer[1] = DATUM_GET_DATA_HWFAB;

	write_dpram_data(buffer, 2);

	read_dpram_data(buffer, 0, 1);

	if (buffer[0] != DATUM_GET_DATA_HWFAB)
		return (EXIT_FAILURE);

	read_dpram_data((char *) &hwfab_number, 1, 2);

	/* Put in host order */

	hwfab_number = ntohs(hwfab_number);

	sprintf(hwfab, "%d", hwfab_number);

	return (EXIT_SUCCESS);
}


/******************* bcmd_get_serial() ******************
*
* Issue command to get board serial number.
*
*/

static int
bcmd_get_serial(char *serial)
{
	uint8_t         buffer[DATUM_MAX_CMD_LEN];
	uint32_t        serial_no;


	buffer[0] = DATUM_GET_DATA_CMD;
	buffer[1] = DATUM_GET_DATA_SERIAL;
	write_dpram_data(buffer, 2);
	read_dpram_data(buffer, 0, 1);
	if (buffer[0] != DATUM_GET_DATA_SERIAL)
		return (EXIT_FAILURE);

	read_dpram_data((char *) &serial_no, 1, 4);

	/* Put in host order */
	serial_no = ntohl(serial_no);

	sprintf(serial, "%d", serial_no);

	return (EXIT_SUCCESS);
}


/******************* bcmd_get_firmware_part() ******************
*
* Issue command to get firmware_part number.
*
*/

static int
bcmd_get_firmware_part(char *firmware_part)
{
	uint8_t         buffer[DATUM_MAX_CMD_LEN];
	uint8_t         major;
	uint8_t         minor;
	uint8_t         month;
	uint8_t         day;
	uint16_t        year;

	buffer[0] = DATUM_GET_DATA_CMD;
	buffer[1] = DATUM_GET_DATA_FWVER;
	write_dpram_data(buffer, 2);
	read_dpram_data(buffer, 0, 1);

	if (buffer[0] != DATUM_GET_DATA_FWVER)
		return (EXIT_FAILURE);

	read_dpram_data((char *) &major, 1, 1);
	read_dpram_data((char *) &minor, 2, 1);
	read_dpram_data((char *) &month, 3, 1);
	read_dpram_data((char *) &day, 4, 1);
	read_dpram_data((char *) &year, 5, 2);
	year = ntohs(year);

	sprintf(firmware_part, "%d.%d %02d/%02d/%04d",
		major, minor, month, day, year);

	return (EXIT_SUCCESS);
}

/******************* bcmd_software_reset() ******************
*
* Issue soft reset command to bc635pci board
*/
static int
bcmd_software_reset(void)
{
	uint8_t         buffer[DATUM_MAX_CMD_LEN];

	buffer[0] = DATUM_SOFT_RESET_CMD;
	write_dpram_data(buffer, 1);
	return (EXIT_SUCCESS);
}

/************** write_dpram_data() **************
*
* Write data to dpram data in area. Low-level routine called only
* by 'bcmd_' functions.
*/

static void
write_dpram_data(char *data, int len)
{
	volatile char  *dpram_addr;

	dpram_addr = glbl.dpram_in_area;

	for (; len > 0; len--) {
		*(dpram_addr++) = *(data++);
	}

	/* Now, jiggle the ack register bit */

	glbl.register_base->ack = DATUM_ACK_BIT_REC;
	glbl.register_base->ack = DATUM_ACK_BIT_CMD;

	/* And wait for the bc635 to execute command */

	while (!(glbl.register_base->ack & DATUM_ACK_BIT_REC))
		usleep(1000);
}

/************** read_dpram_data() **************
 *
 * Read data from dpram data out area. Low-level routine called only
 * by 'xxx' functions.
 */

static void
read_dpram_data(char *data, int offset, int len)
{
	volatile char  *dpram_addr;

	dpram_addr = glbl.dpram_out_area + offset;

	for (; len > 0; len--) {
		*(data++) = *(dpram_addr++);
	}
}





/**************** Interrupt Code ****************
 *
 * Supports only periodic interrupt.
 *
 * Define:
 *
 *   - const struct sigevent isrevent
 *   - handler
 *   - thread function
 *   - function to start up
 *
 ************************************************/

static struct sigevent isrevent;
static const struct sigevent *isr_handler(void *arg, int id);
static void    *int_thread(void *arg);
static int      icount = 0;


/*
 *   Set up interrupt stuff. Called from main().
 */

static int
isr_settup(void)
{

	/*
	 * Make sure interrupts are off to start. software reset does not
	 * affect mask register!
	 */

	glbl.register_base->mask = 0;

	/* Initialize event returned by isr_handler() */

	SIGEV_INTR_INIT(&isrevent);

	/* Create thread to attach to and wait for interrupt */

	pthread_create(NULL, NULL, int_thread, NULL);
}


/*
 * Function to run in thread to wait for interrupt and reply to threads
 * waiting for periodic interrupt.
 */

static void    *
int_thread(void *arg)
{
	int             iid;
	time_struct_t   ts;

	/* Initialize waiting thread structure */

	bc635_waiters_init();

	ThreadCtl(_NTO_TCTL_IO, NULL);

	/*
	 * Interrupts are enables and disabled via io_devctl() application
	 * calls
	 */

	printf("int_thread() attaching interrupt to isr_handler().\n");

	iid = InterruptAttach(glbl.intno, isr_handler, NULL, 0, 0);

	while (1) {
		InterruptWait(NULL, NULL);

		/* Reply to all waiting receivers */

		bc635_waiters_reply(EOK);
	}
	InterruptDetach(iid);
}

/*
 * The interrupt handler itself. This runs at kernel level!
 * Caution!
 */

static const struct sigevent *
isr_handler(void *arg, int id)
{
	/*
	 * Check that one of bc635 interrupts occured. Exit with NULL if not.
	 */

	if (0 == glbl.register_base->intstat & 0x1f)
		return (NULL);

	/*
	 * Reset all interrupts
	 */

	glbl.register_base->intstat = 0x17;
	return (&isrevent);
}

/*
 * Initialize array of rcvid's of endpoiunts waiting for periodic interrupt
 * to complete.
 */

static void
bc635_waiters_init(void)
{
	int             i;

	/* Default attributes are applied */

	pthread_mutex_init(&glbl.bc635_waiters.mut, NULL);

	for (i = 0; i < NWAITMAX; i++) {

		/* 0 is non-reply rcvid */

		glbl.bc635_waiters.rcvids[i].id = 0;
		glbl.bc635_waiters.rcvids[i].count_remaining = 0;
	}
}

/*
 * Add a waiting thread to list
 *
 */

static int
bc635_waiters_add(int rcvid, unsigned int count)
{
	int             i;

	pthread_mutex_lock(&glbl.bc635_waiters.mut);

	for (i = 0; i < NWAITMAX; i++) {
		if (glbl.bc635_waiters.rcvids[i].id == 0) {
			glbl.bc635_waiters.rcvids[i].id = rcvid;
			glbl.bc635_waiters.rcvids[i].count_remaining = count;
			break;
		}
	}

	pthread_mutex_unlock(&glbl.bc635_waiters.mut);

	if (i >= NWAITMAX)
		return (EXIT_FAILURE);

	return (EXIT_SUCCESS);
}

/*
 * Decrement count remaining for each waiting thread/process
 * and if 0, reply to thread.
 *
 */

static void
bc635_waiters_reply(int reply)
{

	int             i;

	pthread_mutex_lock(&glbl.bc635_waiters.mut);
	for (i = 0; i < NWAITMAX; i++) {

		/* (These are guesses as to the last 3 args.) */

		if (glbl.bc635_waiters.rcvids[i].id > 0) {
			if ( --glbl.bc635_waiters.rcvids[i].count_remaining <= 0 ) {
				MsgReply(glbl.bc635_waiters.rcvids[i].id, reply, NULL, 0);
				glbl.bc635_waiters.rcvids[i].id = 0;
			}
		}
	}
	pthread_mutex_unlock(&glbl.bc635_waiters.mut);
}
