/* Test bc635pci driver */

/*
 * DISCLAIMER:
 *
 *   THIS SOFTWARE IS SUPPLIED AS-IS BY THE JOHNS HOPKINS UNIVERSITY/APPLIED PHYSICS
 *   LABORATORY (JHU/APL).
 *
 *   WITH RESPECT TO THE INFORMATION, CODE AND/OR EXECUTABLES PROVIDED HEREIN,
 *   JHU/APL MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR IMPLIED,
 *   INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *   FITNESS FOR A PARTICULAR PURPOSE. IN ADDITION, IN NO EVENT SHALL JHU/APL
 *   BE HELD LIABLE FOR ANY DAMAGES WHATSOEVER INCLUDING DIRECT, INDIRECT, INCIDENTAL,
 *   CONSEQUENTIAL, LOSS OF BUSINESS PROFITS OR SPECIAL DAMAGES, EVEN IF
 *   JHU/APL HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <devctl.h>
#include <math.h>
#include <pthread.h>

#include "bc635.h"
#include "Bc635.h"


/* Periods in microseconds for test */

static unsigned int periods[] = {
	1000000,
	750000,
	500000,
	100000,
	 10000
};

/* Thread function */

static void *periodic( void *argv );
static int fd;

int
main(int argc, char *argv)
{
	int             i;
	int rtn;
		int nloops;
	int             nbytes;
	int             lockstatus;
	double          dtime1;
	double          dtime2;
	double          deltasec;
	double          seconds = 100.0;
	unsigned int    divisors[2];
	u_int           info;
	char            str[100];

	printf("bc635pci test suite.\n");

	/* Open the bancomm bc635pci card. */

	fd = open("/dev/bc635pci", O_RDONLY);
	if (-1 == fd) {
		perror("open(\"/dev/bc635pci\"): ");
		exit(1);
	}
	/* Test the bc635 library devctl() cover functions */

	/* Get model number */

	Bc635_get_model(fd, str, 100);

	printf("%30s %s\n", "Model number:", str);

	/* Get serial number */

	Bc635_get_serial(fd, str, 100);

	printf("%30s %s\n", "Serial number:", str);

	/* Get firmware version number */

	Bc635_get_firmware_part(fd, str, 100);

	printf("%30s %s\n", "Firmware part number:", str);

	/* Get assembly number */

	Bc635_get_assembly(fd, str, 100);

	printf("%30s %s\n", "Assembly number:", str);


	/* Get hwfab number */

	Bc635_get_hwfab(fd, str, 100);

	printf("%30s %s\n", "HW fab number:", str);


	printf("Setting mode to freerun.\n");

	Bc635_set_mode(fd, BC635_MODE_FREERUN);

	printf("Setting time to 2200.000 seconds.\n");

	Bc635_set_time(fd, 2200);

	/* Wait for next 1 PPS epoch */

	sleep(2);

	/* Get time value */

	Bc635_get_time(fd, &dtime1, &lockstatus);

	/* Get string description of lockstatus */

	bc635_lockstatus_str(lockstatus, str);

	printf("Seconds from bc635_gettime: %12.6f\nLockstatus = %s\n", dtime1, str);

	printf("Sleeping 100 ms (using PC clock).\n");

	/* Sleep 100 ms */

	usleep(100000);

	/* Get time value again */

	Bc635_get_time(fd, &dtime2, &lockstatus);

	printf("Times with 100 ms delay: T1 = %12.6f T2 = %12.6f Delta = %12.6f\n",
	       dtime1, dtime2, dtime2 - dtime1);

	/**************** bc635_delay() tests ***************/

	printf("Delaying 10000 tics, with each tic being 100 microseconds.\n");

	Bc635_set_periodic_output(fd, 1000, BC635_SYNC_PPS);

	sleep(1);

	Bc635_enable_int(fd);

	/* Delay one tic to get synced. */

	Bc635_delay( fd, 1);

	Bc635_get_time(fd, &dtime2, &lockstatus);

	printf("Time before delay: %12.6f\n", dtime2);

	Bc635_delay( fd, 1000 );

	Bc635_get_time(fd, &dtime2, &lockstatus);

	printf("Time after delay: %12.6f\n", dtime2);
	
	Bc635_disable_int(fd);


	/**************** Periodic output tests ****************/

	/* Cycle through several periods to test */

	printf("Testing bc635_set_periodic_output() for several periods.\n");

	for (i = 0; i < sizeof(periods) / sizeof(unsigned int); i++) {

		int j;

		/* Set nloops for constant 4 seconds */

		nloops = 4 * (1000000 / periods[i]);

		printf("Setting period to %010d microseconds, with 1 PPS sync.\n", periods[i]);

		Bc635_set_periodic_output(fd, periods[i], BC635_SYNC_PPS);

		printf("Calling bc635_enable_int() for 3 seconds\n");

		Bc635_enable_int(fd);

		for ( j = 0; j < nloops; j++) {
			Bc635_delay( fd, 1 );
			Bc635_get_time(fd, &dtime2, &lockstatus);
			printf("Tic. Time = %12.6f.\n", dtime2);
		}


		printf("Calling bc635_disable_int()\n");

		Bc635_disable_int(fd);

		printf("Setting period to %010d microseconds, without 1 PPS sync.\n", periods[i]);

		Bc635_set_periodic_output(fd, periods[i], BC635_NO_SYNC_PPS);

		printf("Calling bc635_enable_int() for 3 seconds\n");

		Bc635_enable_int(fd);

		for ( j = 0; j < nloops; j++) {
			Bc635_delay( fd, 1 );
			Bc635_get_time(fd, &dtime2, &lockstatus);
			printf("Tic. Time = %12.6f.\n", dtime2);
		}


		printf("Calling bc635_disable_int()\n");

		Bc635_disable_int(fd);

	}


	/* Set up 20 Hz period and start threads waiting for tics. */

	printf("Setting period to 50000 microseconds.\n");

	Bc635_set_periodic_output(fd, 50000, BC635_SYNC_PPS);

	/* Enable interrupts */

	printf("Enabling interrupts.\n");

	Bc635_enable_int(fd);

	printf("Starting %d threads to wait for clock tics (interrupts).\n", 4);

	/* Start threads to wait for tics */

	pthread_create( NULL, NULL, periodic, (void *)1);
	pthread_create( NULL, NULL, periodic, (void *)2);
	pthread_create( NULL, NULL, periodic, (void *)3);
	pthread_create( NULL, NULL, periodic, (void *)4);

	printf("Letting threads run for %d seconds, then turning off interrupts.\n", 5);

	sleep( 5 );

	Bc635_disable_int( fd );

	/**************** End periodic output tests ****************/

	printf("Starting loop to get time with back-to-back calls. Hit ctrl-c to exit...\n");

	while (1) {
		Bc635_get_time(fd, &dtime1, &lockstatus);
		Bc635_get_time(fd, &dtime2, &lockstatus);

		printf(" Times with  Zero delay: T1 = %12.6f T2 = %12.6f Delta = %12.6f\n",
		       dtime1, dtime2, dtime2 - dtime1);
		sleep(1);
	}

}

/* 
 * Function to run as thread to wait for tics. 
 */

static void *periodic( void *argv )
{
	double dtime;
	int lockstatus;

	while (1) {
		
		if ( 0 != bc635_delay(fd, 1)) {
			printf("Thread %3d: bc635_delay() failed. Probably stopped interrupts. Exiting\n",
				(int)argv);
			pthread_exit(NULL);
		}
		Bc635_get_time(fd, &dtime, &lockstatus);
		printf("Thread %3d: %12.6f\n", (int)argv, dtime);
	}
}
		
